<?php

	if (!defined('IN_PARTS')) {
		exit;
	}


	abstract class Filter {
		protected $parameter;
		protected $name;

		private $title;
		private $num_arguments;
		private $arguments = array();


		public function __construct($name, $title, $num_arguments) {
			$this->name = $name;
			$this->title = $title;
			$this->num_arguments = $num_arguments;
		}

		public function setParameter($parameter) {
			$this->parameter = $parameter;
		}

		public function name() {
			return $this->name;
		}


		public function activate() {
			$this->parameter->setCurrentFilter($this->name);
		}


		public function setTitle($title) {
			$this->title = $title;
		}

		public function title() {
			return $this->title;
		}



		public function numArguments() {
			return $this->num_arguments;
		}

		public function setArgument($i, $value) {
			$this->arguments[$i] = $value;
		}

		public abstract function parseArgument($i, $value);

		public function argument($i) {
			return $this->arguments[$i];
		}

		public abstract function formatArgument($i);


		public abstract function buildSql();
	}


	abstract class Parameter {
		protected $module;
		protected $name;

		private $title;
		private $suffixes = array();
		private $term = '';
		private $editable = false;
		private $filters = array();

		private $max_num_arguments = 0;
		private $current_filter = NULL;


		public function __construct($name, $title) {
			$this->name = $name;
			$this->title = $title;

			$this->createFilters();
		}

		protected abstract function createFilters();

		public function setModule($module) {
			$this->module = $module;
		}

		public function name() {
			return $this->name;
		}



		public function setTitle($title) {
			$this->title = $title;
		}

		public function title() {
			return $this->title;
		}


		public function setSuffixes() {
			$this->suffixes = array();

			for ($i = 0; $i < func_num_args(); $i++) {
				$this->suffixes[] = func_get_arg($i);
			}
		}

		public function suffixes() {
			return $this->suffixes;
		}

		public function suffix() {
			if (count($this->suffixes) > 0) {
				return $this->suffixes[0];
			}
			else {
				return '';
			}
		}

		public function hasSuffix() {
			return (count($this->suffixes) > 0);
		}


		public function setTerm($term) {
			$this->term = $term;
		}

		public function term() {
			return $this->term;
		}

		public function buildSql() {
			return '(' . $this->term . ')' . ' AS `P_' . $this->name . '`';
		}

		public function buildSqlSort() {
			return '`P_' . $this->name . '`';
		}



		public function setEditable($editable) {
			$this->editable = $editable;
		}

		public function editable() {
			return $this->editable;
		}


		public function sortable() {
			return true;
		}



		public function maxNumArguments() {
			return $this->max_num_arguments;
		}


		protected function addFilter($filter) {
			$filter->setParameter($this);
			$this->filters[$filter->name()] = $filter;

			$this->max_num_arguments = max(
				$this->max_num_arguments,
				$filter->numArguments()
			);
		}

		public function hasFilters() {
			return count($this->filters);
		}

		public function hasFilter($name) {
			return isset($this->filters[$name]);
		}

		public function &filters() {
			return $this->filters;
		}

		public function filter($name) {
			return $this->filters[$name];
		}

		public function setCurrentFilter($name) {
			$this->current_filter = $this->filters[$name];
			return $this->current_filter;
		}

		public function currentFilter() {
			return $this->current_filter;
		}


		public abstract function sqlParse($row);

		protected abstract function parse($string, &$value);

		public function parseValue($string, &$value) {
			return $this->parse($string, $value);
		}


		public abstract function sqlFormat($value);

		protected abstract function format($value);

		public function formatField($row) {
			return $this->format($this->sqlParse($row));
		}

		public function formatValue($value) {
			return $this->format($value);
		}
	}


	abstract class Module {
		private $name;
		private $parameters = array();
		private $sql_columns = '';

		private static $instances = array();

		public static function instance($name, $arguments) {
			if (isset(self::$instances[$name])) {
				return self::$instances[$name];
			}

			require_once 'modules/' . strtolower($name) . '.php';

			$class = $name . 'Module';
			$object = new $class($arguments);

			$object->name = strtolower($name);

			self::$instances[$name] = $object;
			return $object;
		}



		protected function addParameter($parameter) {
			$parameter->setModule($this);
			$this->parameters[$parameter->name()] = $parameter;

			/* Add to SQL query */
			$this->sql_columns .=
				(($this->sql_columns == '') ? '' : ', ') .
				$parameter->buildSql();
		}

		public function hasParameter($name) {
			return isset($this->parameters[$name]);
		}

		public function &parameters() {
			return $this->parameters;
		}

		public function parameter($name) {
			return $this->parameters[$name];
		}


		public function buildSqlColumns() {
			return $this->sql_columns;
		}

		public function buildSqlFilters() {
			$sql = '';
			foreach ($this->parameters as $parameter) {
				if ($parameter->currentFilter() === NULL) {
					continue;
				}

				$sql .= ' AND ' . $parameter->currentFilter()->buildSql();
			}

			return substr($sql, 5);
		}

		public function buildSqlTable() {
			return '`' . $this->name . '_models`';
		}

		public function buildSqlConditions() {
			return '';
		}
	}






	class NativeFilter extends Filter {
		private $sql_template = '';

		public function __construct($name, $title, $template, $num_arguments) {
			parent::__construct($name, $title, $num_arguments);
			$this->sql_template = $template;
		}

		public function parseArgument($i, $value) {
			/* Default is to use native format */
			$result = NULL;
			if (!$this->parameter->parseValue($value, $result)) {
				return false;
			}

			$this->setArgument($i, $result);
			return true;
		}

		public function formatArgument($i) {
			/* Default is to use native format */
			return $this->parameter->formatValue($this->argument($i));
		}

		public function buildSql() {
			$arguments = array(
				'X' => $this->parameter->term()
			);

			for ($i = 0; $i < $this->numArguments(); $i++) {
				$arguments[strval($i + 1)] =
					$this->parameter->sqlFormat($this->argument($i));
			}

			return '(' . keysprintf($this->sql_template, $arguments) . ')';
		}
	}


	class RangeFilter extends Filter {
		private $sql_template;

		public function __construct($name, $title, $template) {
			parent::__construct($name, $title, 1);
			$this->sql_template = $template;
		}


		public function parseArgument($i, $value) {
			/* Default is to use native format */
			$result = NULL;
			if (!$this->parameter->parseFilterValue($value, $result)) {
				return false;
			}

			$this->setArgument($i, $result);
			return true;
		}

		public function formatArgument($i) {
			/* Default is to use native format */
			return $this->parameter->formatFilterValue($this->argument($i));
		}


		public function buildSql() {
			$arguments = array(
				'L' => $this->parameter->sqlLowBound(),
				'H' => $this->parameter->sqlHighBound(),
				'1' => $this->parameter->sqlFilterFormat($this->argument(0))
			);

			return '(' . keysprintf($this->sql_template, $arguments) . ')';
		}
	}






	class DecimalParameter extends Parameter {
		private $precision = NULL;

		public function __construct($name, $title, $precision = NULL) {
			parent::__construct($name, $title);
			$this->precision = $precision;
		}

		protected function createFilters() {
			$this->addFilter(new NativeFilter('gt',
				'ist größer als', '%X > %1', 1));

			$this->addFilter(new NativeFilter('gte',
				'ist größer als oder gleich', '%X >= %1', 1));

			$this->addFilter(new NativeFilter('eq',
				'ist gleich', 'ABS(%X - %1) < 1', 1));

			$this->addFilter(new NativeFilter('lte',
				'ist kleiner als oder gleich', '%X <= %1', 1));

			$this->addFilter(new NativeFilter('lt',
				'ist kleiner als', '%X < %1', 1));

			$this->addFilter(new NativeFilter('neq',
				'ist ungleich', 'ABS(%X - %1) > 1', 1));

			$this->addFilter(new NativeFilter('bt',
				'liegt zwischen', '%1 BETWEEN %1 AND %2', 2));
		}


		public function sqlParse($row) {
			global $Db;
			return $Db->parseDecimal($row['P_' . $this->name]);
		}

		protected function parse($string, &$value) {
			/* Get magnitude and remove suffix */
			$preg_suffixes = '';
			foreach ($this->suffixes() as $suffix) {
				$preg_suffixes .= preg_quote($suffix) . '|';
			}

			$result = preg_match(
				'/^([\\s0-9\\.,+-]+)([afpnumkGTE]?)\\s*' .
					'(' . substr($preg_suffixes, 0, -1) . ')?\\s*$/i',
				$string,
				$matches
				);

			if ($result == 0) {
				/* Cancel */
				return false;
			}

			/*
				Array
				(
						[0] => -123.987  G OHM
						[1] => -123.987
						[2] => G
						[3] => OHM
				)
			*/

			$value = parse_decimal($matches[1]);

			if ($matches[2] != '') {
				$magnitude = strpos('afpnum kMGTPE', $matches[2]) - 6;
				$value *= pow(1000, $magnitude);
			}

			return true;
		}


		public function sqlFormat($value) {
			global $Db;
			return $Db->formatDecimal($value);
		}

		protected function format($value) {
			$magnitude = calculate_magnitude($value);
			$result = format_decimal($value, $this->precision);

			if ( ($magnitude != '') || $this->hasSuffix() ) {
				$result .= ' ';
			}

			return $result . $magnitude . $this->suffix();
		}
	}


	class DecimalRangeParameter extends DecimalParameter {
		private $sql_tolerance;


		protected function createFilters() {
			$this->addFilter(new RangeFilter('con',
				'enthält', '%1 BETWEEN %L AND %H'));

			$this->addFilter(new RangeFilter('ncon',
				'enthält nicht', '%1 NOT BETWEEN %L AND %H'));
		}

		public function sortable() {
			return false;
		}



		public function setTolerance($tolerance) {
			$this->sql_tolerance = $tolerance;
		}

		public function tolerance() {
			return $this->sql_tolerance;
		}


		public function sqlLowBound() {
			return '(' .
				$this->term() .
				' * (1 - (' . $this->sql_tolerance . ' / 100)))';
		}

		public function sqlHighBound() {
			return '(' .
				$this->term() .
				' * (1 + (' . $this->sql_tolerance . ' / 100)))';
		}

		public function buildSql() {
			return
				$this->sqlLowBound() . ' AS `Pl_' . $this->name . '`, ' .
				$this->sqlHighBound() . ' AS `Ph_' . $this->name . '`';
		}




		public function sqlParse($row) {
			global $Db;
			return array(
				$Db->parseDecimal($row['Pl_' . $this->name]),
				$Db->parseDecimal($row['Ph_' . $this->name])
			);
		}

		protected function parse($string, &$value) {
			return false;
		}


		public function sqlFormat($value) {
			return NULL;
		}

		protected function format($value) {
			return
				parent::format($value[0]) .
				' … ' . /* Unicode */
				parent::format($value[1]);
		}




		public function parseFilterValue($string, &$value) {
			return parent::parse($string, $value);
		}

		public function formatFilterValue($value, $fancy = false) {
			return parent::format($value, $fancy);
		}

		public function sqlFilterFormat($value) {
			return parent::sqlFormat($value);
		}
	}


	class StringParameter extends Parameter {
		protected function createFilters() {
			$this->addFilter(new NativeFilter('bg',
				'beginnt mit', '%X LIKE CONCAT(%1, \'%%\')', 1));

			$this->addFilter(new NativeFilter('nbg',
				'beginnt nicht mit', '%X NOT LIKE CONCAT(%1, \'%%\')', 1));

			$this->addFilter(new NativeFilter('ed',
				'endet mit', '%X LIKE CONCAT(\'%%\', %1)', 1));

			$this->addFilter(new NativeFilter('ned',
				'endet nicht mit', '%X LIKE CONCAT(\'%%\', %1)', 1));

			$this->addFilter(new NativeFilter('con',
				'enthält', '%X LIKE CONCAT(\'%%\', %1, \'%%\')', 1));

			$this->addFilter(new NativeFilter('ncon',
				'enthält nicht', '%X NOT LIKE CONCAT(\'%%\', %1, \'%%\')', 1));

			$this->addFilter(new NativeFilter('em',
				'ist leer', '%X = \'\'', 0));

			$this->addFilter(new NativeFilter('nem',
				'ist nicht leer', '%X <> \'\'', 0));
		}



		public function sqlParse($row) {
			global $Db;
			return $Db->parseString($row['P_' . $this->name]);
		}

		protected function parse($string, &$value) {
			$value = strtolower(trim($string));

			/* Find suffix */
			foreach ($this->suffixes() as $suffix) {
				$suffix_length = strlen($suffix);
				if (substr($value, (-1) * $suffix_length)
					== strtolower($suffix)) {

					$value = substr($value, 0, (-1) * $suffix_length);
					break;
				}
			}

			return true;
		}



		public function sqlFormat($value) {
			global $Db;
			return $Db->formatString($value);
		}

		protected function format($value) {
			return
				$value .
				(($this->hasSuffix()) ? ' ' : '') .
				$this->suffix();
		}
	}


	class BooleanParameter extends Parameter {
		protected function createFilters() {
			$this->addFilter(new NativeFilter('t',
				'ist wahr', '%X == \'1\'', 0));

			$this->addFilter(new NativeFilter('f',
				'ist falsch', '%X == \'0\'', 0));
		}


		public function sqlFormat($value) {
			return '\'' . (($value) ? '1' : '0') . '\'';
		}

		protected function parse($string, &$value) {
			$string = strtolower(trim($string));

			/* Suffix doesn't make sense here */

			static $true_values = array(
				'1', '+',
				'y', 'yes', 't', 'true',
				'j', 'ja', 'w', 'wahr'
			);

			static $false_values = array(
				'0', '-',
				'n', 'no', 'f', 'false',
				'nein', 'falsch'
			);

			if (in_array($string, $true_values)) {
				$value = true;
				return true;
			}
			elseif (in_array($string, $false_values)) {
				$value = false;
				return true;
			}

			return false;
		}


		public function sqlParse($row) {
			return ($row['P_' . $this->name] == '1');
		}

		protected function format($value) {
			return (($value) ? 'wahr' : 'falsch');
		}
	}





?>